// Copyright (C) 2013 S&P Capital IQ
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.spcapitaliq.platformexecuting.sbtplugin

import java.util.regex.Matcher.quoteReplacement

import sbt._
import Keys._
import Project.Initialize
import scalaz.syntax.apply.^
import scalaz.syntax.semigroup._
import scalaz.syntax.std.option._
import scalaz.std.option._
import scalaz.std.string._
import com.spcapitaliq.platformexecuting.{Capture, Platform}

/** SBT setting keys for capturing platform-specific binaries. */
sealed abstract class PESpecSettingKeys {
  import PESpec._

  /** Platform IDs and locations of executables to capture. */
  val binaries: SettingKey[Binaries]

  /** Version of the underlying package. */
  val version: SettingKey[String]

  /** Basename of exe, e.g. "phantomjs". */
  val executable: SettingKey[String]

  /** Where to store in classpath, e.g. "com/example/my/bins". */
  val resourceRoot: SettingKey[String]

  /** Environment variable name to override stored binaries. */
  val environmentVar: SettingKey[String]

  /** Additional licenses that apply to the included binaries. */
  val binaryLicenses: SettingKey[Seq[(String, URL)]]

  /** Fully-qualified class name to generate. */
  val javaFQClassName: SettingKey[String]
}

/** SBT settings for capturing platform-specific binaries.
  *
  * @param sbtId Prefix for newly introduced sbt setting keys.
  */
final case class PESpec(sbtId: String) extends PESpecSettingKeys {
  import PESpec._

  val binaries = SettingKey[Binaries](
    "%s-binaries" format sbtId,
    "Platform IDs and locations of %s executables to capture." format sbtId)

  val version = SettingKey[String]("%s-version" format sbtId, "%s version" format sbtId)

  val executable = SettingKey[String](
    "%s-executable" format sbtId,
    "Basename of the %s executable, e.g. \"phantomjs\"." format sbtId)

  val resourceRoot = SettingKey[String](
    "%s-resource-root" format sbtId,
    "Where to store in classpath, e.g. \"com/example/my/bins\".")

  val environmentVar = SettingKey[String](
    "%s-environment-var" format sbtId,
    "Environment variable name to override stored binaries.")

  val binaryLicenses = SettingKey[Seq[(String, URL)]](
    "%s-binary-licenses" format sbtId,
    "Additional licenses that apply to the included binaries.")

  val javaFQClassName = SettingKey[String](
    "%s-java-fq-class-name" format sbtId,
    "Fully-qualified class name to generate.")

  /** A best-effort guess as to what binaries to include.  If you want
    * portable jars, you should do something different.
    *
    * @todo pick the platform of the exe, not *my* platform (e.g. exe
    *       is x86, I run on amd64)
    */
  lazy val defaultBinaries: Initialize[Binaries] =
    (executable, environmentVar){(exe, env) =>
      ^(Platform.platformSpec,
        Capture.localExe(exe, Some(env)).toOption){
        (ps, ex) => Map(ps -> ex)
      } getOrElse Map()
    }

  /** Include these settings in your sbt project where you want to
    * package the binaries.
    */
  lazy val settings: Seq[Setting[_]] = Seq(
    // Include binaries in binary package.
    mappings in (Compile, packageBin) <++=
      (streams, binaries, executable, resourceRoot) map {(s, b, exe, resroot) =>
        b.toSeq map {
          case (plat, fil) =>
            if (!fil.isFile())
              s.log.warn("%s is missing and will not be packaged" format fil)
            fil -> Capture.captureTo(exe = exe, resourceRoot = resroot, plat)
        }
      }

   ,licenses <++= (binaries, binaryLicenses)((b, bs) => if (b.isEmpty) Seq.empty else bs)

    // Add a source file publishing the same exe, resourceRoot, and
    // env that we used in generating this binary.
   ,sourceGenerators in Compile <+=
      (javaFQClassName, executable, resourceRoot, environmentVar,
       sourceManaged in Compile) map {(fqcn, exe, rsrt, env, to: File) =>
        val (pn, cn) = splitJC(fqcn)
        writeSpecIdJava(pn, cn, exe, rsrt, env, to)
      }

    // don't add _2.10 to artifact name
   ,crossVersion := CrossVersion.Disabled)
}

object PESpec {
  type Binaries = Map[Platform, File]

  private def splitJC(fqcn: String): (String, String) =
    (psplit findFirstMatchIn fqcn
       cata (m => (m group 1, m group 2),
             sys error ("Bad classname " |+| fqcn)))

  private val psplit = """(?s)(.+)\.(.+)""".r

  private val tmplvar = """@([^@]+)@""".r

  private lazy val specidtmpl = getClass getResource "PESpecIds.java.in"

  private[sbtplugin]
  def writeSpecIdJava(javaClassPackageName: String, javaClassBaseName: String,
                      exe: String, resourceRoot: String, env: String,
                      todir: File): Seq[File] = {
    val exp = Map[String, String](
      "JAVACLASSPACKAGENAME" -> javaClassPackageName,
      "JAVACLASSBASENAME" -> javaClassBaseName,
      "EXE" -> exe,
      "RESOURCEROOT" -> resourceRoot,
      "ENV" -> env)
    def exptmpl(s: String): String =
      tmplvar replaceAllIn (s, m => quoteReplacement(exp(m.group(1))))
    val of = todir / (javaClassBaseName |+| ".java")
    IO.writeLines(of, IO.readLinesURL(specidtmpl).view map exptmpl)
    Seq(of)
  }
}
